"""block_cluster_tree.py: :class:`BlockClusterTree`,
:func:`build_block_cluster_tree`,
:func:`recursion_build_block_cluster_tree`
"""
from HierMat.cluster_tree import admissible, ClusterTree


class BlockClusterTree(object):
    """Compares two cluster trees level wise with respect to an admissibility condition and builds a tree
    """
    def __init__(self, left_clustertree, right_clustertree, sons=None, level=0, is_admissible=False, plot_info=None):
        # type: (ClusterTree, ClusterTree, list, int, bool, tuple) -> None
        self.sons = sons if sons else []
        self.left_clustertree = left_clustertree
        self.right_clustertree = right_clustertree
        self.admissible = is_admissible
        self.level = level
        self.plot_info = plot_info

    def __repr__(self):
        optional_string = " with children {0!s}".format(self.sons) if self.sons else ""
        return "<BlockClusterTree at level {0}{1}>".format(str(self.level), optional_string)

    def __len__(self):
        left_len = len(self.left_clustertree)
        right_len = len(self.right_clustertree)
        return left_len * right_len

    def __eq__(self, other):
        """Test for equality
        :type other: BlockClusterTree
        """
        return (self.left_clustertree == other.left_clustertree
                and self.right_clustertree == other.right_clustertree
                and self.sons == other.sons
                and self.admissible == other.admissible
                and self.level == other.level
                and self.plot_info == other.plot_info
                )

    def __ne__(self, other):
        return not self == other

    def depth(self, root_level=None):
        """Get depth of the tree

        :param root_level: internal use.
        :type root_level: int
        :return: depth
        :rtype: int

        .. hint::

            **recursive function**
        """
        if root_level is None:
            root_level = self.level
        if not self.sons:
            return self.level - root_level
        else:
            return max([son.depth(root_level) for son in self.sons])

    def to_list(self):
        """Give list representation for export

        Return a list containing the object and a list with its sons

        .. hint::

            **recursive function**
        """
        return [self, [son.to_list() for son in self.sons]]

    def shape(self):
        """Return length of left and right cluster tree

        :return: x and y dimension
        :rtype: tuple(int, int)
        """
        return len(self.left_clustertree), len(self.right_clustertree)

    def draw(self, axes, admissible_color='#1e26bc', inadmissible_color='#bc1d38'):
        """Draw a patch into given axes

        :param axes: axes instance to draw in
        :type axes: matplotlib.pyplot.axes
        :param admissible_color: color for admissible patch (see matplotlib for color specs)
        :type admissible_color: str
        :param inadmissible_color: color for inadmissible patch
        :type inadmissible_color: str
        """
        # set x coordinates for patch
        x_min, y_min = self.plot_info
        x_max = x_min + len(self.left_clustertree)
        y_max = y_min + len(self.right_clustertree)
        x = [x_min, x_min, x_max, x_max]
        # set y coordinates for patch
        y = [y_min, y_max, y_max, y_min]
        color = admissible_color if self.admissible else inadmissible_color
        axes.fill(x, y, color, ec='k', lw=0.1)

    def plot_recursion(self, axes, admissible_color='#1e26bc', inadmissible_color='#bc1d38'):
        if self.sons:
            for son in self.sons:
                son.plot_recursion(axes, admissible_color=admissible_color, inadmissible_color=inadmissible_color)
        else:
            self.draw(axes, admissible_color=admissible_color, inadmissible_color=inadmissible_color)

    def to_xml(self):
        """Return a string for xml representation"""
        out_list = self.to_list()
        return self._to_xml(out_list)

    def to_dot(self):
        """Return a string for .dot representation"""
        out_list = self.to_list()
        return self._to_dot(out_list)

    @staticmethod
    def _to_xml(lst, out_string=''):
        if len(lst[1]):
            value_string = str(lst[0])
            display_string = str(len(lst[1]))
        else:
            value_string = str(lst[0])
            display_string = str(lst[0])
        out_string += '<node value="{0}">{1}\n'.format(value_string, display_string)
        if len(lst) > 1 and type(lst[1]) is list:
            for item in lst[1]:
                out_string = BlockClusterTree._to_xml(item, out_string)
        out_string += "</node>\n"
        return out_string

    @staticmethod
    def _to_dot(lst, out_string=''):
        if len(lst) > 1:
            for item in lst[1]:
                if type(item) is list:
                    value_string = str(lst[0])
                    item_string = str(item[0])
                    label_string = len(lst[0])
                    out_string += '''"{0}" -- "{1}";
                            "{0}"[label="{2}",color="#cccccc",style="filled",shape="box"];\n'''.format(
                        value_string, item_string, label_string)
                    out_string = BlockClusterTree._to_dot(item, out_string)
        return out_string


def build_block_cluster_tree(left_cluster_tree, right_cluster_tree=None, start_level=0, admissible_function=admissible):
    """Factory for building a block cluster tree

    :param left_cluster_tree: "left side" cluster tree
    :type left_cluster_tree: ClusterTree
    :param right_cluster_tree: "right side" cluster tree
    :type right_cluster_tree: ClusterTree
    :param start_level: counter that keeps track of the level
    :type start_level: int
    :param admissible_function: function that determines whether two cluster trees are admissible or not. This is
        crucial for the structure of the block cluster tree. (See :mod:`utils.admissible` for an example)
    :type admissible_function: function that returns a bool
    :return: block cluster tree
    :rtype: BlockClusterTree
    """
    if not right_cluster_tree:
        right_cluster_tree = left_cluster_tree
    is_admissible = admissible_function(left_cluster_tree, right_cluster_tree)
    x_min, x_max = left_cluster_tree.get_patch_coordinates()
    y_min, y_max = right_cluster_tree.get_patch_coordinates()
    plot_info = (x_min, y_min)
    root = BlockClusterTree(left_cluster_tree, right_cluster_tree,
                            level=start_level, is_admissible=is_admissible, plot_info=plot_info)
    recursion_build_block_cluster_tree(root, admissible_function)
    return root


def recursion_build_block_cluster_tree(current_tree, admissible_function):
    """Recursion to :func:`build_block_cluster_tree`
    """
    if not admissible_function(current_tree.left_clustertree, current_tree.right_clustertree):
        # get top left corner of current block
        x_min, y_min = current_tree.plot_info
        x_current = x_min
        y_current = y_min
        for left_son in current_tree.left_clustertree.sons:
            for right_son in current_tree.right_clustertree.sons:
                new_tree = BlockClusterTree(left_son, right_son,
                                            level=current_tree.level + 1,
                                            is_admissible=False,
                                            plot_info=(x_current, y_current)
                                            )
                current_tree.sons.append(new_tree)
                recursion_build_block_cluster_tree(new_tree, admissible_function)
                y_current += len(right_son)
            y_current = y_min
            x_current += len(left_son)
    else:
        current_tree.admissible = True



"""utils.py: Utilities for the :mod:`HMatrix` module
"""
import math

from HierMat.block_cluster_tree import BlockClusterTree
from HierMat.cluster_tree import ClusterTree
from HierMat.grid import Grid


def export(obj, form='xml', out_file='out'):
    """Export obj in specified format.
    
    :param obj: object to export
    :type obj: BlockClusterTree or ClusterTree or HMat or RMat
    :param form: format specifier
    :type form: str
    :param out_file: path to output file
    :type out_file: str
    :raises NotImplementedError: if form is not supported

    .. note::

        implemented so far:

        - xml
        - dot
        - bin
    """
    if form == 'xml':
        head = '<?xml version="1.0" encoding="utf-8"?>\n'
        output = obj.to_xml()
        output = head + output
        with open(out_file, "w") as out:
            out.write(output)
    elif form == 'dot':
        head = 'graph {\nnodesep=0.1;\nranksep=1.5;\n'
        output = obj.to_dot()
        tail = '}'
        output = head + output + tail
        with open(out_file, "w") as out:
            out.write(output)
    elif form == 'bin':
        import pickle
        file_handle = open(out_file, "wb")
        pickle.dump(obj, file_handle, protocol=-1)
        file_handle.close()
    else:
        raise NotImplementedError()


def plot(obj, filename=None, **kwargs):
    """plot an object
    
    :param obj: object to plot
    :type obj: BlockClusterTree or Grid
    :param filename: filename to save the plot to (if omitted, the plot will be displayed)
    :type filename: str
    :param kwargs: optional arguments to specific plot commands
        see the respective documentations
    """
    if isinstance(obj, BlockClusterTree):
        return block_cluster_tree_plot(obj, filename, **kwargs)
    else:
        raise NotImplementedError('object can not be plotted')


def block_cluster_tree_plot(obj, filename=None, ticks=False, face_color='#133f52',
                            admissible_color='#76f7a8', inadmissible_color='#ff234b'):
    """Plot the block cluster tree
    
    :param obj: block cluster tree to plot
    :type obj: BlockClusterTree
    :param filename: filename to save the plot. if omitted, the plot will be displayed
    :type filename: str
    :param ticks: show ticks in the plot
    :type ticks: bool
    :param face_color: background color (see matplotlib for color specs)
    :param admissible_color: color for admissible patch
    :type admissible_color: str
    :param inadmissible_color: color for inadmissible patch
    :type inadmissible_color: str

    .. note::

        depends on :mod:`matplotlib.pyplot`

    """
    import matplotlib.pyplot as plt

    plt.rc('axes', linewidth=0.5, labelsize=4)
    plt.rc('xtick', labelsize=4)
    plt.rc('ytick', labelsize=4)
    fig = plt.figure(figsize=(3, 3), dpi=400)
    fig.patch.set_facecolor(face_color)
    # get max of the ticks
    x_min, x_max = obj.left_clustertree.get_patch_coordinates()
    y_min, y_max = obj.right_clustertree.get_patch_coordinates()
    axes = plt.axes()
    axes.set_xlim(x_min, x_max + 1)
    axes.set_ylim(y_min, y_max + 1)
    if ticks:
        x_divisors = list(divisor_generator(x_max + 1))
        y_divisors = list(divisor_generator(y_max + 1))
        if len(x_divisors) > 4:
            x_ticks = x_divisors[-4]
        else:
            x_ticks = x_divisors[-1]
        if len(y_divisors) > 4:
            y_ticks = y_divisors[-4]
        else:
            y_ticks = y_divisors[-1]
        axes.set_xticks(range(x_min, x_max + 2, x_ticks))
        axes.set_yticks(range(y_min, y_max + 2, y_ticks))
    else:
        axes.set_xticks([])
        axes.set_yticks([])
    axes.tick_params(length=2, width=0.5)
    axes.xaxis.tick_top()
    axes.invert_yaxis()
    obj.plot_recursion(axes, admissible_color=admissible_color, inadmissible_color=inadmissible_color)
    fig.add_axes(axes)
    if not filename:
        return fig
    else:
        # remove whitespace around the plot
        plt.subplots_adjust(left=0, right=1, top=1, bottom=0)
        plt.savefig(filename, format='png', facecolor=fig.get_facecolor(), edgecolor=None)


def load(filename):
    """Load a :class:`ClusterTree` or :class:`BlockClusterTree` from file

    :param filename: file to import
    :type filename: String
    :return: object
    :rtype: BlockClusterTree or ClusterTree

    .. note:: Depends on :mod:`pickle`
    """
    import pickle
    with open(filename, 'rb') as infile:
        obj = pickle.load(infile)
    return obj


def divisor_generator(n):
    """Return divisors of n

    :param n: integer to find divisors of
    :type n: int
    :return: divisors
    :rtype: list[int]

    .. warning::
       This is a generator! To get a list with all divisors call::

          list(divisor_generator(n))


    .. note::
       found at
       `StackOverflow
       <http://stackoverflow.com/questions/171765/what-is-the-best-way-to-get-all-the-divisors-of-a-number>`_
       on 2017.03.08
    """
    large_divisors = []
    for i in xrange(1, int(math.sqrt(n) + 1)):
        if n % i == 0:
            yield i
            if i * i != n:
                large_divisors.append(n / i)
    for divisor in reversed(large_divisors):
        yield divisor



"""cluster.py: :class:`Cluster` object and iterator
"""
from HierMat.grid import Grid

import numpy


class Cluster(object):
    """Handles operations on a :class:`Grid` object by manipulating an index list
    """

    def __init__(self, grid, indices=None):
        """Create a cluster

        :param grid: grid to build cluster around
        :type grid: Grid
        :param indices: index list (optional)
        :type indices: list
        """
        self.grid = grid
        self.indices = range(len(grid)) if not indices else indices

    def __getitem__(self, item):
        """Get item

        :param item: index
        :type item: int
        """
        return self.grid[self.indices[item]]

    def __repr__(self):
        """String representation
        :rtype: str
        """
        return "<Cluster object with grid {0} and indices {1}>".format(self.grid, self.indices)

    def __iter__(self):
        """Iterate through Cluster
        """
        return ClusterIterator(self)

    def __len__(self):
        """Number of points
        """
        return len(self.indices)

    def __eq__(self, other):
        """Test for equality
        """
        return self.grid == other.grid and self.indices == other.indices

    def __ne__(self, other):
        """Test for inequality
        """
        return not self == other

    def get_grid_item(self, item):
        """Return item from grid

        :param item: index
        :type item: int
        """
        return self.grid.get_point(item)

    def get_grid_item_support_by_index(self, item):
        """Return supports of i-th item from grid

        :param item: index
        :type item: int
        """
        return self.grid.get_support_by_index(item)

    def get_grid_item_support(self, item):
        """Return supports of item from grid

        :param item: point
        :type item: tuple(float)
        """
        return self.grid.get_support(item)

    def get_index(self, item):
        """Return index at item

        :param item: index
        :type item: int
        :return: item-th index
        :rtype: int
        """
        return self.indices[item]

    def get_patch_coordinates(self):
        """Return min and max out of indices

        :return: min and max
        :rtype: tuple(int, int)
        """
        return min(self.indices), max(self.indices)

    def dim(self):
        """Compute dimension

        :return: dim of Grid
        :rtype: int
        """
        return self.grid.dim()

    def diameter(self):
        """Compute diameter

        Return the maximal Euclidean distance between two points
        For big Clusters this is costly

        :return: diameter
        :rtype: float
        """
        # get all points from grid in indices
        points = [self.grid.points[i] for i in self.indices]
        # compute distance matrix
        dist_mat = [numpy.linalg.norm(numpy.array(x) - numpy.array(y)) for x in points for y in points]
        return max(dist_mat)

    def distance(self, other):
        """Compute distance to other cluster

        Return minimal Euclidean distance between points of self and other

        :param other: another instance of Cluster

        :return: distance
        :rtype: float
        """
        return min([numpy.linalg.norm(x - y) for x in self for y in other])


class ClusterIterator(object):
    """Iterator to Cluster object"""

    def __init__(self, cluster):
        self.cluster = cluster
        self._counter = 0

    def __iter__(self):
        return self

    def next(self):
        if self._counter >= len(self.cluster):
            raise StopIteration
        else:
            self._counter += 1
            return self.cluster[self._counter - 1]



"""rmat.py: :class:`RMat`
"""
import numbers

import numpy


class RMat(object):
    """Rank-k matrix

    Implementation of the low rank matrix as described in [HB2015]
    
    .. note:
        There is still an issue with __radd__, when the left operand is a numpy.ndarray object
    """

    def __init__(self, left_mat, right_mat=None, max_rank=None):
        """Build Rank-k matrix

        Arguments:
            left_mat: numpy.matrix n x k
            right_mat: numpy.matrix m x k
            max_rank: integer max_rank
        """
        # check for wrong input
        if right_mat is None and max_rank is None:
            raise ValueError('Not enough input arguments. At least one of right_mat and max_rank has to be specified')
        if max_rank is not None and max_rank <= 0:
            raise ValueError('max_rank must be a positive integer')

        # check for direct conversion
        if right_mat is None:
            self._from_matrix(left_mat, max_rank)
        else:  # default constructor
            left_shape = left_mat.shape
            right_shape = right_mat.shape
            if left_shape[1] != right_shape[1]:
                raise ValueError('shapes {0.shape} and {1.shape} not aligned: '
                                 '{0.shape[1]} (dim 1) != {1.shape[1]} (dim 1)'.format(left_mat, right_mat))
            self.max_rank = max_rank
            self.left_mat = left_mat
            self.right_mat = right_mat
            self.shape = (left_shape[0], right_shape[0])
            # check if we have to reduce
            if self.max_rank and self.max_rank < left_shape[1]:
                self._reduce(self.max_rank)

    @staticmethod
    def from_matrix(matrix, max_rank):
        """Convert a full matrix to the rank k format 
        
        :param matrix: Matrix to convert to RMat
        :type matrix: numpy.matrix
        :param max_rank: maximum rank
        :type max_rank: int
        :return: Best approximation of matrix in RMat format
        :rtype: RMat
        """
        u, s, v = numpy.linalg.svd(matrix)
        left_mat = u[:, 0: max_rank] * numpy.diag(s[0: max_rank])
        right_mat = v[0: max_rank, :].T
        return RMat(left_mat, right_mat, max_rank)

    def _from_matrix(self, matrix, max_rank):
        """Convert a full matrix to the rank k format 

        :param matrix: Matrix to convert to RMat
        :type matrix: numpy.matrix
        :param max_rank: maximum rank
        :type max_rank: int
        """
        self.max_rank = max_rank
        u, s, v = numpy.linalg.svd(matrix)
        self.left_mat = u[:, 0: self.max_rank] * numpy.diag(s[0: self.max_rank])
        self.right_mat = v[0: self.max_rank, :].T
        self.shape = matrix.shape

    def __repr__(self):
        left_str = self.left_mat.__repr__().replace('\n', '').replace(' ', '')
        right_str = self.right_mat.__repr__().replace('\n', '').replace(' ', '')
        out_str = '<RMat with left_mat: {0}, right_mat: {1} and max_rank: {2}>'.format(left_str,
                                                                                       right_str,
                                                                                       self.max_rank)
        return out_str

    def __str__(self):
        left_str = str(self.left_mat)
        right_str = str(self.right_mat)
        out_str = 'Rank-k matrix with left block:\n{0}\nand right block:\n{1}'.format(left_str, right_str)
        return out_str

    def __eq__(self, other):
        """Check for equality"""
        left_eq = numpy.array_equal(self.left_mat, other.left_mat)
        right_eq = numpy.array_equal(self.right_mat, other.right_mat)
        return left_eq and right_eq and self.max_rank == other.max_rank

    def __ne__(self, other):
        return not self == other

    def __add__(self, other):
        """Addition of self and other"""
        try:
            if self.shape != other.shape:
                raise ValueError('operands could not be broadcast together with shapes '
                                 '{0.shape} {1.shape}'.format(self, other))
        except AttributeError:
            if not isinstance(other, numbers.Number):
                raise NotImplementedError('unsupported operand type(s) for +: {0} and {1}'.format(type(self),
                                                                                                  type(other)))
            else:
                return self + numpy.matrix(other)
        if isinstance(other, RMat):
            # if max_rank is defined do exact, else do exact
            if self.max_rank:
                return self.form_add(other, self.max_rank)
            else:
                return self._add_rmat(other)
        elif isinstance(other, numpy.matrix):
            return self._add_matrix(other)
        else:
            raise NotImplementedError('unsupported operand type(s) for +: {0} and {1}'.format(type(self), type(other)))

    def _add_rmat(self, other):
        """Add two Rank-k-matrices

        Resulting matrix will have higher rank
        """
        new_left = numpy.concatenate([self.left_mat, other.left_mat], axis=1)
        new_right = numpy.concatenate([self.right_mat, other.right_mat], axis=1)
        return RMat(new_left, new_right)

    def _add_matrix(self, other):
        """Add full matrix
        
        :type other: numpy.matrix
        """
        # TODO: check if suitable
        addend = self.from_matrix(other, self.max_rank)
        return self + addend

    def __radd__(self, other):
        """Should be commutative so just switch"""
        return self + other

    def __sub__(self, other):
        """Subtract two Rank-k-matrices"""
        return self + (-other)

    def __neg__(self):
        """Unary minus"""
        new_k = self.max_rank
        new_left = numpy.matrix(-self.left_mat)
        new_right = numpy.matrix(self.right_mat)
        return RMat(new_left, new_right, new_k)

    def __abs__(self):
        """Frobenius-norm"""
        return numpy.linalg.norm(self.left_mat * self.right_mat.T)

    def norm(self, order=None):
        """Norm of the matrix

        :param order: order of the norm (see in :func:`numpy.linalg.norm`)
        :return: norm
        :rtype: float
        """
        return numpy.linalg.norm(self.left_mat * self.right_mat.T, ord=order)

    def split(self, start_x, start_y, end_x, end_y):
        """Return the block of self that matches the supplied indices
        
        :param start_x: x start index
        :type start_x: int
        :param start_y: y start index
        :type start_y: int
        :param end_x: x end index
        :type end_x: int
        :param end_y: y end index
        :type end_y: int
        :return: rmat that represents the corresponding block
        :rtype: RMat
        """
        return RMat(left_mat=self.left_mat[start_x: end_x, :],
                    right_mat=self.right_mat[start_y: end_y, :],
                    max_rank=self.max_rank)

    def __mul__(self, other):
        """Multiplication of self and other"""
        if isinstance(other, RMat):
            return self._mul_with_rmat(other)
        elif isinstance(other, numpy.matrix):
            return self._mul_with_mat(other)
        elif isinstance(other, numpy.ndarray):
            return self._mul_with_vector(other)
        elif isinstance(other, numbers.Number):
            return self._mul_with_int(other)
        else:
            raise NotImplementedError('unsupported operand type(s) for *: {0} and {1}'.format(type(self), type(other)))

    def _mul_with_mat(self, other):
        """Multiplication with full matrix"""
        j, r = self.right_mat.shape
        j2, r2 = other.shape
        if j != j2:
            raise ValueError('shapes {0.shape} and {1.shape} not aligned: '
                             '{0.shape[1]} (dim 1) != {1.shape[0]} (dim 0)'.format(self, other))
        return RMat(self.left_mat, other.T * self.right_mat, r)

    def _mul_with_rmat(self, other):
        """Multiplication with rmat"""
        i, r1 = self.left_mat.shape
        j, r1 = self.right_mat.shape
        j2, r2 = other.left_mat.shape
        k, r2 = other.right_mat.shape
        out_rank = max((self.max_rank, other.max_rank))
        if j != j2:
            raise ValueError('shapes {0.shape} and {1.shape} not aligned: '
                             '{0.shape[1]} (dim 1) != {1.shape[0]} (dim 0)'.format(self, other))
        cost1 = 2 * r1 * r2 * (i + j) - r2 * (i + r1)
        cost2 = 2 * r1 * r2 * (j + k) - r1 * (k + r2)
        if cost2 >= cost1:
            return RMat(self.left_mat * (self.right_mat.T * other.left_mat), other.right_mat, out_rank)
        else:
            return RMat(self.left_mat, other.right_mat * (other.left_mat.T * self.right_mat), out_rank)

    def _mul_with_vector(self, other):
        """Multiplication with vector"""
        return self.left_mat * (self.right_mat.T * other)

    def _mul_with_int(self, other):
        """Multiplication with scalar"""
        return RMat(other * self.left_mat, self.right_mat, self.max_rank)

    def __rmul__(self, other):
        """Multiplication numpy.matrix * RMat"""
        if isinstance(other, numpy.matrix):
            # return RMat(other * self.left_mat, self.right_mat, self.max_rank)
            return other * self.to_matrix()
        elif isinstance(other, numbers.Number):
            return self._mul_with_int(other)
        else:
            raise NotImplementedError('unsupported operand type(s) for *: {0} and {1}'.format(type(self), type(other)))

    def form_add(self, other, rank=None):
        """Formatted addition of self and other, i.e. addition and reduction to rank::

            (self + other).reduce(rank)

        If rank is omitted, reduction to min(rank(self), rank(other))

        :param other: other rank-k matrix
        :type other: RMat
        :param rank: rank after reduction
        :type rank: int
        :return: reduced result
        :rtype: RMat
        """
        if not rank:
            rank = min((self.max_rank, other.max_rank))
        res = self._add_rmat(other)
        res._reduce(rank)
        return res

    def to_matrix(self):
        """Full matrix representation::

            left_mat * right_mat.T

        :return: full matrix
        :rtype: numpy.matrix
        """
        return self.left_mat * self.right_mat.T

    def reduce(self, new_k):
        """Perform a reduced QR decomposition to rank new_k

        :param new_k: rank to reduce to
        :type new_k: int
        :return: reduced matrix
        :rtype: RMat
        """
        q_left, r_left = numpy.linalg.qr(self.left_mat)
        q_right, r_right = numpy.linalg.qr(self.right_mat)
        temp = r_left * r_right.T
        u, s, v = numpy.linalg.svd(temp)
        new_left = q_left * u[:, 0:new_k] * numpy.diag(s[0:new_k])
        new_right = q_right * v[:, 0:new_k]
        # noinspection PyTypeChecker
        return RMat(new_left, new_right, new_k)

    def _reduce(self, new_k):
        """Perform a reduced QR decomposition to rank new_k internal

        :param new_k: rank to reduce to
        :type new_k: int
        """
        q_left, r_left = numpy.linalg.qr(self.left_mat)
        q_right, r_right = numpy.linalg.qr(self.right_mat)
        temp = r_left * r_right.T
        u, s, v = numpy.linalg.svd(temp)
        new_left = q_left * u[:, 0:new_k] * numpy.diag(s[0:new_k])
        new_right = q_right * v[:, 0:new_k]
        self.left_mat = new_left
        self.right_mat = new_right
        self.max_rank = new_k

    def transpose(self):
        """Return transposed copy of self
        
        :rtype: RMat
        """
        out_max_rank = self.max_rank
        out_left = self.right_mat
        out_right = self.left_mat
        return RMat(left_mat=out_left, right_mat=out_right, max_rank=out_max_rank)
